####################################################################################################
# Inverted Likelihood Ratio R Script
####################################################################################################
#
# This file provides the main function to calculate confidence sets for break dates as 
# documented in:
#
# Yunjong Eo and James Morley (2015)
#   "Likelihood-Ratio-Based Confidence Sets for the Timing of Structural Breaks",
#   Quantitative Economics, Volume 6, Issue 2, pages 463–-497, DOI: 10.3982/QE186.
#
# These codes are a direct translation of the original GAUSS procs written by:
# Yunjong Eo and James Morley (2015)
#
# R emulation of GAUSS procs, printing and plotting codes by Luke Hartigan, 28-07-2016
####################################################################################################

# Source utility functions
source("invLR_utils.R")

invLR_cs <- function(y, z, S, m, bigt, R, ...) {
  UseMethod("invLR_cs")
}

invLR_cs.default <- function(y, z, S, m, bigt, br, hetq, vauto, brv, brbeta, R, prewhit, trm) {

    # Get some useful quantities
    y <- as.matrix(y)
    z <- as.matrix(z)
    n <- ncol(y)
    
    # Set up regression matrices
    tmp <- transf(y, z, S)
    maty <- tmp$maty
    matz <- tmp$matz
    
    # Estimate the restricted model
    tmp <- restim(maty, matz, bigt, n, m, br, R, brbeta, brv)
    rbetast <- tmp$nbeta
    rvvst <- tmp$nvv
    likst <- estim_lik(maty, matz, n, m, br, rbetast, rvvst)
    
    # Compute scale factors
    wmat <- ws(maty, matz, m, n, bigt, br, rbetast, rvvst, hetq, vauto, brv, brbeta, prewhit)
    
    kmat <- matrix(data = 0.0, nrow = m, ncol = 3L)
    
    j <- 1L
    while (j <= m) {
        w1 <- wmat[j, 1L]
        w2 <- wmat[j, 2L]
        
        kmat[j, 1L] <- cv(w1, w2, 0.99)
        kmat[j, 2L] <- cv(w1, w2, 0.95)
        kmat[j, 3L] <- cv(w1, w2, 0.90)
        
        j <- j + 1L
    }
    
    csm99 <- csm95 <- csm90 <- matrix(data = NA_real_, nrow = bigt, ncol = m)
    
    gp <- round(trm * bigt)
    
    brm <- rbind(0, br, nrow(maty)/n)
    
    k0 <- 1L
    
    while (k0 <= m) {
        i <- brm[k0] + gp
        j <- brm[k0 + 2L] - gp
        
        cn <- i
        while (cn <= j) {
            brc <- br
            brc[k0] <- cn
            tmp <- restim(maty, matz, bigt, n, m, brc, R, brbeta, brv)
            nbeta <- tmp$nbeta
            nvv <- tmp$nvv
            lik <- estim_lik(maty, matz, n, m, brc, nbeta, nvv)
            
            if ((likst - lik) <= kmat[k0, 1L]) {
                csm99[cn, k0] <- 1L
            }
            
            if ((likst - lik) <= kmat[k0, 2L]) {
                csm95[cn, k0] <- 1L
            }
            
            if ((likst - lik) <= kmat[k0, 3L]) {
                csm90[cn, k0] <- 1L
            }
            
            cn <- cn + 1L
        }
        
        k0 <- k0 + 1L
    }
    
    # Collect results as a list object
    invlr <- list(y = y, br = br, csm99 = csm99, csm95 = csm95, csm90 = csm90, 
        kmat = kmat, rbetast = rbetast, rvvst = rvvst)
    
    # Return results with class "invLR"
    invlr$call <- match.call()
    class(invlr) <- "invLR"
    return(invlr)
}

# Function to calculate scale factors for the limit distribution
# See Eo and Morley (2015) for notations (B1, B2, Q1, Q2, Pi1, Pi2, Omega1, Omega2, Psi1, Psi2, Gam1, Gam2)
ws <- function(maty, matx, m, n, bigt, br, betax, vv, hetq, vauto, brv, brbeta, prewhit) {
    wmat <- matrix(data = 0.0, nrow = m, ncol = 2)
    diagx <- pzbar(matx, m, br, bigt)
    br <- rbind(br, (nrow(maty) / n))
    res <- maty - (diagx %*% betax)
    res <- reshapex(res, nrow(res) / n, n)
    
    # Get standardised residuals
    j <- 1L
    while (j <= (m + 1)) {
        if (j == 1L) {
            res[1:br[j], ] <- res[1:br[j], ] %*% qr.solve(sqrm(vv[1:(j * n),]))
        } else {
            res[(br[j - 1] + 1):br[j], ] <- res[(br[j - 1] + 1):br[j], ] %*% qr.solve(sqrm(vv[((j - 1) * n + 1):(j * n), ]))
        }
    j <- j + 1L
    }
    
    j <- 1L
    while (j <= m) {
        vvar1 <- vv[((j - 1) * n + 1):(j * n), ]
        vvar2 <- vv[(j * n + 1):((j + 1) * n), ]
        B1 <- sqrm(vvar1) %*% qr.solve(vvar2) %*% (vvar2 - vvar1) %*% qr.solve(sqrm(vvar1))
        B2 <- sqrm(vvar2) %*% qr.solve(vvar1) %*% (vvar2 - vvar1) %*% qr.solve(sqrm(vvar2))
        Q1 <- Q2 <- Pi1 <- Pi2 <- 0.0
        if (j == 1L) {
            i <- 1L
        } else {
            i <- br[j - 1] + 1L
        }
        if (hetq == 0) {
            # Change in variance and/or regression coefficients, 
            # allow for serial correlation and change in the distribution of the regressors
            if ((brv == 1) && (vauto == 1)) { 
                # Construct estimate of Pi1
                tempmat <- matrix(data = 0.0, nrow = (br[j] - i + 1), ncol = ncol(matx))
                k <- 1L
                while (k <= (br[j] - i + 1)) {
                    tempmat[k, ] <- t(t(matx[(((i-1+k-1)*n+1):(i-1+k)*n), , drop = FALSE]) %*% qr.solve(vvar2) %*% sqrm(vvar1) %*% t(res[i-1+k, , drop = FALSE]))
                    k <- k + 1L
                }
                Pi1 <- correct(tempmat, prewhit)        # with scaling
                
                # Construct estimate of Omega1
                tempmat <- matrix(data = 0.0, nrow = (br[j] - i + 1), n^2)
                k <- 1L
                while (k <= (br[j] - i + 1)) {
                    tempmat[k, ] <- t(vec(crossprod(x = res[(i - 1 + k), , drop = FALSE], y = NULL) - eye(n)))
                    k <- k + 1L
                }
                Omega1 <- correct(tempmat, prewhit)     # with scaling
                
                while (i <= br[j]) {
                    Q1 <- Q1 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar2) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    i <- i + 1L
                }
                
                if (j == 1) {
                    Q1 <- Q1 / br[j]
                } else {
                    Q1 <- Q1 / (br[j] - br[j - 1])
                }
                
                i <- br[j] + 1L
                
                # Construct estimate of Pi2
                tempmat <- matrix(data = 0.0, nrow = (br[j + 1] - i + 1), ncol = ncol(matx))
                k <- 1L
                while (k <= (br[j + 1] - i + 1)) {
                    tempmat[k, ] <- t(t(matx[(((i-1+k-1)*n+1):(i-1+k)*n), , drop = FALSE]) %*% qr.solve(vvar1) %*% sqrm(vvar2) %*% t(res[i-1+k, , drop = FALSE]))
                    k <- k + 1L
                }
                Pi2 <-  correct(rempmat, prewhit)       # with scaling
                
                # Construct estimate of Omega2
                tempmat <- matrix(data = 0.0, nrow = (br[j + 1] - i + 1), n^2)
                k <- 1L
                while (k <= (br[j + 1] - i + 1)) {
                    tempmat[k, ] <- t(vec(crossprod(x = res[(i - 1 + k), , drop = FALSE], y = NULL) - eye(n)))
                    k <- k + 1L
                }
                Omega2 <- correct(tempmat, prewhit)     # with scaling
                
                while (i <= br[j + 1]) {
                    Q2 <- Q2 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar1) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    i <- i + 1L
                }
                Q2 <- Q2 / (br[j + 1] - br[j])
                
                diffb <- betax[(j * ncol(matx) + 1):((j + 1) * ncol(matx))] - betax[((j - 1) * ncol(matx) + 1):(j * ncol(matx))]
                Psi1 <- 0.5 * sumc(diag(B1 %*% B1)) + t(diffb) %*% Q1 %*% diffb
                Psi2 <- 0.5 * sumc(diag(B2 %*% B2)) + t(diffb) %*% Q2 %*% diffb
                Gam1 <- sqrt(0.25 * t(vec(B1)) %*% Omega1 %*% vec(B1) + t(diffb) %*% Pi1 %*% diffb)
                Gam2 <- sqrt(0.25 * t(vec(B2)) %*% Omega2 %*% vec(B2) + t(diffb) %*% Pi2 %*% diffb)
            } 
            # Change in the variance and/or in the regression coefficients
            if ((brv == 1) && (vauto == 0)) {
                # Construct estimate of Omega1
                Omega1 <- 0
                while (i <= br[j]) {
                    Omega1 <- Omega1 + tcrossprod(x = vec(crossprod(x = res[i, , drop = FALSE], y = NULL) - eye(n)), y = NULL)
                    Q1 <- Q1 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar2) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    Pi1 <- Pi1 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar2) %*% vvar1 %*% qr.solve(vvar2) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    i <- i + 1L
                }
                
                if (j == 1L) {
                    Omega1 <- Omega1 / br[j]
                    Q1 <- Q1 / br[j]
                    Pi1 <- Pi1 / br[j]
                } else {
                    Omega1 <- Omega1 / (br[j] - br[j - 1])
                    Q1 <- Q1 / (br[j] - br[j - 1])
                    Pi1 <- Pi1 / (br[j] - br[j - 1])
                }
                
                i <- br[j] + 1L
                
                # Construct estimate of Omega2
                Omega2 <- 0
                while (i <= br[j + 1]) {
                    Omega2 <- Omega2 + tcrossprod(x = vec(crossprod(x = res[i, , drop = FALSE], y = NULL) - eye(n)), y = NULL)
                    Q2 <- Q2 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar1) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    Pi2 <- Pi2 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar1) %*% vvar2 %*% qr.solve(vvar1) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    i <- i + 1
                }
                
                Omega2 <- Omega2 / (br[j] - br[j - 1])
                Q2 <- Q2 / (br[j] - br[j - 1])
                Pi2 <- Pi2 / (br[j] - br[j - 1])
                
                diffb <- betax[(j * ncol(matx) + 1):((j + 1) * ncol(matx))] - betax[((j - 1) * ncol(matx) + 1):(j * ncol(matx))]
                Psi1 <- 0.5 * sumc(diag(B1 %*% B1)) + t(diffb) %*% Q1 %*% diffb
                Psi2 <- 0.5 * sumc(diag(B2 %*% B2)) + t(diffb) %*% Q2 %*% diffb
                Gam1 <- sqrt(0.25 * t(vec(B1)) %*% Omega1 %*% vec(B1) + t(diffb) %*% Pi1 %*% diffb)
                Gam2 <- sqrt(0.25 * t(vec(B2)) %*% Omega2 %*% vec(B2) + t(diffb) %*% Pi2 %*% diffb)
            }
            # Change in the coefficients only
            if ((brv == 0) && (brbeta == 1) && (vauto == 1)) {
                # Construct estimate of Pi1
                tempmat <- matrix(data = 0.0, nrow = bigt, ncol = ncol(matx))
                k <- 1L
                while (k <= bigt) {
                    tempmat[k, ] <- t(t(matx[((k - 1)*n+1):(k * n), , drop = FALSE]) %*% qr.solve(vvar2) %*% sqrm(vvar1) %*% t(res[k, , drop = FALSE]))
                    k <- k + 1L
                }
                Pi1 <- correct(tempmat, prewhit)    # with scaling
                k <- 1L
                while (k <= bigt) {
                    Q1 <- Q1 + t(matx[((k - 1)*n+1):(k * n), , drop = FALSE]) %*% qr.solve(vvar2) %*% matx[((k - 1)*n+1):(k * n), , drop = FALSE]
                    k <- k + 1L
                }
                Q1 <- Q1 / bigt
                #Pi2 <- Pi1
                #Q2 <- Q1
                diffb <- betax[(j * ncol(matx) + 1):((j + 1) * ncol(matx))] - betax[((j - 1) * ncol(matx) + 1):(j * ncol(matx))]
                Psi1 <- t(diffb) %*% Q1 %*% diffb
                Psi2 <- Psi1 # t(diffb) %*% Q2 %*% diffb
                Gam1 <- sqrt(t(diffb) %*% Pi1 %*% diffb)
                Gam2 <- Gam1 # Gam1 <- sqrt(t(diffb) %*% Pi2 %*% diffb)
            }
            # Change in the coefficients only
            if ((brv == 0) && (brbeta == 1) && (vauto == 0)) {
                k <- 1L
                while (k <= bigt) {
                    Q1 <- Q1 + t(matx[((k - 1)*n+1):(k * n), , drop = FALSE]) %*% qr.solve(vvar2) %*% matx[((k - 1)*n+1):(k * n), , drop = FALSE]
                    Pi1 + t(matx[((k - 1)*n+1):(k * n), , drop = FALSE]) %*% qr.solve(vvar2) %*% vvar1 %*% qr.solve(vvar2) %*% matx[((k - 1)*n+1):(k * n), , drop = FALSE]
                    k <- k + 1L
                }
                Q1 <- Q1 / bigt
                Pi1 <- Pi1 / bigt
                #Pi2 <- Pi1
                #Q2 <- Q1
                diffb <- betax[(j * ncol(matx) + 1):((j + 1) * ncol(matx))] - betax[((j - 1) * ncol(matx) + 1):(j * ncol(matx))]
                Psi1 <- t(diffb) %*% Q1 %*% diffb
                Psi2 <- Psi1 # t(diffb) %*% Q2 %*% diffb
                Gam1 <- sqrt(t(diffb) %*% Pi1 %*% diffb)
                Gam2 <- Gam1 # Gam1 <- sqrt(t(diffb) %*% Pi2 %*% diffb)
            }
        } else if (hetq == 1) {
            # Allow the distribution of the regressors to be different, we only need to separately treat vauto == [1 || 0],
            # for a partial break model, the corresponding coefficients are zero
            if (vauto == 1) {
                # Construct estimate of Pi1
                tempmat <- matrix(data = 0.0, nrow = (br[j]- i + 1), ncol = ncol(matx))
                k <- 1L
                while (k <= (br[j] - i + 1)) {
                    tempmat <- t(t(matx[((i-1+k-1)*n+1):((i-1+k)*n), , drop = FALSE]) %*% qr.solve(vvar2) %*% sqrm(vvar1) %*% t(res[(i-1+k), , drop = FALSE]))
                    k <- k + 1L
                }
                Pi1 <- correct(tempmat, prewhit)    # with scaling
                
                # Construct estimate of Omega1
                tempmat <- matrix(data = 0.0, nrow = (br[j]-i+1), ncol = n^2)
                k <- 1L
                while (k <= (br[j] - i + 1)) {
                    tempmat[k, ] <- t(vec(crossprod(x = res[(i-1+k), , drop = FALSE], y = NULL) - eye(n)))
                    k <- k + 1L
                }
                Omega1 <- correct(tempmat, prewhit) # with scaling
                while (i <= br[j]) {
                    Q1 <- Q1 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar2) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    i <- i + 1L
                }
                
                if (j == 1L) {
                    Q1 <- Q1 / br[j]
                } else {
                    Q1 <- Q1 / (br[j] - br[j - 1])
                }
                
                i <- br[j] + 1
                
                # Construct estimate of Pi2
                tempmat <- matrix(data = 0.0, nrow = (br[j + 1]-i+1), ncol = ncol(matx))
                k <- 1L
                while (k <= (br[j + 1] - i + 1)) {
                    tempmat[k, ] <- t(t(matx[((i-1+k-1)*n+1):((i-1+k)*n), , drop = FALSE]) %*% qr.solve(vvar1) %*% sqrm(vvar2) %*% t(res[(i-1+k), , drop = FALSE]))
                    k <- k + 1L
                }
                Pi2 <- correct(tempmat, prewhit)    # with scaling
                
                # Construct estimate of Omega2
                tempmat <- matrix(data = 0.0, nrow = (br[j+1]-i+1), ncol = n^2)
                k <- 1L
                while (k <= (br[j + 1] - i + 1)) {
                    tempmat[k, ] <- t(crossprod(x = res[(i-1+k), , drop = FALSE], y = NULL) - eye(n))
                    k <- k + 1L
                }
                Omega2 <- correct(tempmat, prewhit) # with scaling
                while (i <= br[j + 1]) {
                    Q2 <- Q2 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar1) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    i <- i + 1L
                }
                Q2 <- Q2 / (br[j + 1] - br[j])
                diffb <- betax[(j * ncol(matx) + 1):((j + 1) * ncol(matx))] - betxa[((j - 1) * ncol(matx) + 1):(j * ncol(matx))]
                Psi1 <- 0.5 * sumc(diag(B1 %*% B1)) + t(diffb) %*% Q1 %*% diffb
                Psi2 <- 0.5 * sumc(diag(B2 %*% B2)) + t(diffb) %*% Q2 %*% diffb
                Gam1 <- sqrt(0.25 * t(vec(B1)) %*% Omega1 %*% vec(B1) + t(diffb) %*% Pi1 %*% diffb)
                Gam2 <- sqrt(0.25 * t(vec(B2)) %*% Omega2 %*% vec(B2) + t(diffb) %*% Pi2 %*% diffb)
            }
            # Change in the variance and/or in the regression coefficients
            if (vauto == 0) {
                # Construct estimate of Omega1
                Omega1 <- 0
                while (i <= br[j]) {
                    Omega1 <- Omega1 + tcrossprod(x = vec(crossprod(x = res[i, , drop = FALSE], y = NULL) - eye(n)), y = NULL)
                    Q1 <- Q1 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar2) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    Pi1 <- Pi1 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar2) %*% vvar1 %*% qr.solve(vvar2) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    i <- i + 1L
                }
                
                if (j == 1) {
                    Omega1 <- Omega1 / br[j]
                    Q1 <- Q1 / br[j]
                    Pi1 <- Pi1 / br[j]
                } else {
                    Omega1 <- Omega1 / (br[j] - br[j - 1])
                    Q1 <- Q1 / (br[j] - br[j - 1])
                    Pi1 <- Pi1 / (br[j] - br[j - 1])
                }
                
                i <- br[j] + 1
                
                # Construct estimate of Omega2
                Omega2 <- 0
                while (i <= br[j + 1]) {
                    Omega2 <- Omega2 + tcrossprod(x = vec(crossprod(x = res[i, , drop = FALSE], y = NULL) - eye(n)), y = NULL)
                    Q2 <- Q2 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar1) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    Pi2 <- Pi2 + t(matx[((i - 1)*n+1):(i * n), , drop = FALSE]) %*% qr.solve(vvar1) %*% vvar2 %*% qr.solve(vvar1) %*% matx[((i - 1)*n+1):(i * n), , drop = FALSE]
                    i <- i + 1L
                }
                
                Omega2 <- Omega2 / (br[j + 1] - br[j])
                Q2 <- Q2 / (br[j + 1] - br[j])
                Pi2 <- Pi2 / (br[j + 1] - br[j])
                
                diffb <- betax[(j * ncol(matx) + 1):((j + 1) * ncol(matx))] - betax[((j - 1) * ncol(matx) + 1):(j * ncol(matx))]
                Psi1 <- 0.5 * sumc(diag(B1 %*% B1)) + t(diffb) %*% Q1 %*% diffb
                Psi2 <- 0.5 * sumc(diag(B2 %*% B2)) + t(diffb) %*% Q2 %*% diffb
                Gam1 <- sqrt(0.25 * t(vec(B1)) %*% Omega1 %*% vec(B1) + t(diffb) %*% Pi1 %*% diffb)
                Gam2 <- sqrt(0.25 * t(vec(B2)) %*% Omega2 %*% vec(B2) + t(diffb) %*% Pi2 %*% diffb)
            }
        }        
        
        w1 <- (Gam1^2) / Psi1
        w2 <- (Gam2^2) / Psi2
        wmat[j, ] <- cbind(w1, w2)   
        j <- j + 1L
    }
    
    # Return the scale factors
    return (wmat)
}

# Function to calculate the critical value for the inverted LR confidence set
cv <- function(w1, w2, lv, max_iter = 1000) {
    
    a <- 1 - lv
    w <- rbind(w1, w2)
    
    uw <- maxc(w)
    lw <- minc(w)
    
    uk <- -uw * log(1 - sqrt(lv))
    lk <- -lw * log(1 - sqrt(lv))
    
    k <- 0.5 * (uk + lk)  
    cr <- (1.0 - exp(-k / w1)) * (1.0 - exp(-k / w2))
    
    itr <- 1L
    while ((abs(lv - cr) >= 1.0E-5) && (itr <= max_iter)) {
        if ((cr - lv) > 0.0 ) {
            uk <- k
        } else {
            lk <- k
        }
        
        uc <- (1.0 - exp(-uk / w1)) * (1.0 - exp(-uk / w2))
        lc <- (1.0 - exp(-lk / w1)) * (1.0 - exp(-lk / w2))
        
        k <- ((lv - lc) * uk + (uc - lv) * lk) / (uc - lc)
        cr <- (1.0 - exp(-k / w1)) * (1.0 - exp(-k / w2))
        
        itr <- itr + 1L
        if (itr == max_iter) {
            warning("Iteration has reached the upper bound for finding cv of invLR.\n", call. = FALSE)
            break
        }
    }
    return (k)
}   

# Function to compute the log of the likelihood value given the break dates
estim_lik <- function(maty, matz, n, m, br, betax, vv) {
    
    qn <- ncol(matz)
    lik <- 0.0
    br <- rbind(0, br, (nrow(maty) / n))
    k <- 1L
    while (k <= m + 1L) {
        i <- br[k] + 1L
        j <- br[k + 1L]
        segx <- matz[((i - 1L) * n + 1L):(j * n), , drop = FALSE]
        segy <- maty[((i - 1L) * n + 1L):(j * n), , drop = FALSE]
        b <- betax[(qn * (k - 1L) + 1L):(qn * k), 1L]
        vvar <- vv[(n * (k - 1L) + 1L):(n * k), 1L:n]
        res <- segy - segx %*% b
        umat <- reshapex(res, nrow(res) / n, n)
        
        if (is.null(dim(vvar))) {
            lik <- lik - ((j - i + 1L) * n * 0.5 * log(2.0 * pi)) - (j - i + 1L)  * 0.5 * log(vvar) - 0.5 * 
                sumc(sumc(umat %*% qr.solve(vvar) * umat))
        } else {
            lik <- lik - ((j - i + 1L) * n * 0.5 * log(2.0 * pi)) - (j - i + 1L)  * 0.5 * log(det(vvar)) - 0.5 * 
                sumc(sumc(umat %*% qr.solve(vvar) * umat))
        }
        
        k <- k + 1L
    }
    return (lik)
}   

# EOF
