####################################################################################################
# Inverted Likelihood Ratio Test
####################################################################################################
#
# This file is used to test the "invLR.R" codes.
#
# Yunjong Eo and James Morley (2015)
#   "Likelihood-Ratio-Based Confidence Sets for the Timing of Structural Breaks",
#   Quantitative Economics, Volume 6, Issue 2, pages 463–-497, DOI: 10.3982/QE186.
#
# Luke Hartigan, 28-07-2016
####################################################################################################

# Clear the workspace before beginning
rm(list = ls(all = TRUE))

# Source the required functions
source("invLR.r")

# Set up a few R options
options(digits = 4)

# Select which replication to run
Model <- c("GDP", "CON", "YC")
id <- 3

# Set up the model to test
if (Model[id] == "GDP") {

# What are we doing?
cat("\nAR(1) Model for U.S. Output Growth in Section 4 of EM(2015)...\n")

# Read in the test data and do transformations
ycdata <- read.csv("ycdata.csv", header = TRUE)
ycdata <- 400.0 * log(ycdata)
y <- as.matrix(ycdata[3:nrow(ycdata), 1] - ycdata[2:(nrow(ycdata)-1), 1])
z <- as.matrix(cbind(1.0, (ycdata[2:(nrow(ycdata)-1), 1] - ycdata[1:(nrow(ycdata)-2), 1])))

# Specify some function options
S <- eye(2)
m <- 1
bigt <- nrow(y)
R <- eye(ncol(S) * (m + 1))
trm <- 0.15
brv <- 1
brbeta <- 1
vauto <- 0
prewhit <- 0
hetq <- 1
br <- 144

} else if (Model[id] == "CON") {

# What are we doing?
cat("\nAR(2) Model for U.S. Consumption Growth in Section 4 of EM(2015)...\n")

# Read in the test data and do transformations
ycdata <- read.csv("ycdata.csv", header = TRUE)
ycdata <- 400.0 * log(ycdata)
y <- as.matrix(ycdata[4:nrow(ycdata), 2] - ycdata[3:(nrow(ycdata)-1), 2])
z <- as.matrix(cbind(1.0, 
              (ycdata[3:(nrow(ycdata)-1), 2] - ycdata[2:(nrow(ycdata)-2), 2]),
              (ycdata[2:(nrow(ycdata)-2), 2] - ycdata[1:(nrow(ycdata)-3), 2])))


# Specify some function options
S <- eye(3)
m <- 2
bigt <- nrow(y)
R <- eye(ncol(S) * (m + 1))
trm <- 0.15
brv <- 1
brbeta <- 1
vauto <- 0
prewhit <- 0
hetq <- 1
br <- c(44, 184)
br <- make_br_vec(br) # NB: 'br' must be a column vector for multiple breaks

} else if (Model[id] == "YC") {

# What are we doing?
cat("\nVECM(2) Model for U.S. Output-Consumption Growth in Section 4 of EM(2015)...\n")

# Read in the test data and do transformations
ycdata <- read.csv("ycdata.csv", header = TRUE)
ycdata <- 400.0 * log(ycdata)
y <- as.matrix(ycdata[4:nrow(ycdata), ] - ycdata[3:(nrow(ycdata)-1), ])
n <- ncol(y)
z <- as.matrix(cbind(1.0, 
              (ycdata[3:(nrow(ycdata)-1), 2] - ycdata[3:(nrow(ycdata)-1), 1]),
              (ycdata[3:(nrow(ycdata)-1), ] - ycdata[2:(nrow(ycdata)-2), ]),
              (ycdata[2:(nrow(ycdata)-2), ] - ycdata[1:(nrow(ycdata)-3), ])))

# Specify some function options
S <- eye(n * ncol(z))
m <- 4
bigt <- nrow(y)
R <- eye(ncol(S) * (m + 1))
trm <- 0.15
brv <- 1
brbeta <- 1
vauto <- 0
prewhit <- 0
hetq <- 1
br <- c(56, 101, 141, 194)
br <- make_br_vec(br) # NB: 'br' must be a column vector for multiple breaks

} else {

stop("Invalid option.\n\n", call. = FALSE)

}

# Run the code for the specified model and options provided
results <- invLR_cs(y, z, S, m, bigt, br, hetq, vauto, brv, brbeta, R, prewhit, trm)

# Print output to the console
print(results, print_all = FALSE)

# Do some plots
plot(results)

# EOF
