####################################################################################################
# Inverted Likelihood Ratio Utility Functions
####################################################################################################
#
# This file provides utility functions to calculate confidence sets for break dates as 
# documented in:
#
# Yunjong Eo and James Morley (2015)
#   "Likelihood-Ratio-Based Confidence Sets for the Timing of Structural Breaks",
#   Quantitative Economics, Volume 6, Issue 2, pages 463–-497, DOI: 10.3982/QE186.
#
# These codes are a direct translation of the original GAUSS procs written by:
# Qu & Perron (2007). 
#
# NB: We made one change from the GAUSS code in the computation of the Andrews (1991) automatic 
# bandwidth function. We use <x,y> / <x,x> instead of using an explicit regression function.
#
# R emulation of GAUSS procs, printing and plotting codes by Luke Hartigan, 28-07-2016
####################################################################################################

# R emulation of GAUSS eye(n)
eye <- function(n) {
    return (diag(x = 1, nrow = n))
}

# R emulation of GAUSS maxc(x)
maxc <- function(x) {
    return (matrix(data = apply(as.matrix(x), 2, max), ncol = 1))
}

# R emulation of GAUSS minc(x)
minc <- function(x) {
    return (matrix(data = apply(as.matrix(x), 2, min), ncol = 1))
}

# R emulation of GAUSS sumc(x)
sumc <- function(x) {
    return (matrix(data = apply(as.matrix(x), 2, sum), ncol = 1))
}    

# R emulation of GAUSS vec(x)
vec <- function(x) {
    return (matrix(data = as.matrix(x), nrow = length(x), ncol = 1))
}

# R emulation of GAUSS vecr(x)
vecr <- function(x) {
    return (matrix(data = t(as.matrix(x)), nrow = length(x), ncol = 1))
}

# R emulation of GAUSS reshape(x, r, c)
reshapex <- function(x, m, n) {

    # Get some useful quantities
    x <- as.matrix(x)
    i <- nrow(x)
    j <- ncol(x)

    size_x <- i * j
    size_y <- m * n
    
    if (size_x == size_y) {
        y <- vec(t(x))
        y <- matrix(data = y, nrow = m, ncol = n, byrow = TRUE)
    } else if(size_x < size_y) {
        y <- rep(vec(t(x)), times = size_y)
        y <- matrix(data = y, nrow = m, ncol = n, byrow = TRUE)
    } else {
        y <- vec(t(x))
        y <- matrix(data = y[1:size_y], nrow = m, ncol = n, byrow = TRUE)
    } 
    return (y)
}

# Transform dependent and explanatory variables for estimation
transf <- function(y, z, S) {
    
    # Calculate a few useful quantities
    y <- as.matrix(y)
    z <- as.matrix(z)
    m <- nrow(y)
    n <- ncol(y)
    ID <- eye(n)
    
    maty <- t(y[1, , drop = FALSE])
    matz <- ID %x% z[1, , drop = FALSE] %*% S
 
    i <- 2L
    while (i <= m) {
        maty <- rbind(maty, t(y[i, , drop = FALSE]))
        matz <- rbind(matz, ID %x% z[i, , drop = FALSE] %*% S)
        i <- i + 1L
    }
    return (list(maty = maty, matz = matz))
}

# Compute the square root of a symmetric positive definition matrix
sqrm <- function(m) {
    # check dimensions
    if (is.null(dim(m))) {
        sqm <- sqrt(m)
    } else {
        tmp <- eigen(m, symmetric = TRUE)
        D <- diag(x = tmp$values, nrow = nrow(m))
        V <- tmp$vectors
        sqm <- V %*% sqrt(D) %*% t(V)
    }
    return (sqm)
}

# Estimate the coefficients under the optimal partitions
estim <- function(maty, matz, n, m, br, max_iter = 200) {
    bhat <- matrix(data = 0.0, nrow = (m + 1), ncol = ncol(matz))
    vv <- matrix(data = 0.0, nrow = ((m + 1) * n), ncol = n)
    br <- rbind(0, br, (nrow(maty) / n))
    k <- 1L
    while (k <= (m + 1)) {
        i <- br[k] + 1
        j <- br[k + 1]
        segy <- maty[((i - 1)*n+1):(j * n), , drop = FALSE]
        segx <- matz[((i - 1)*n+1):(j * n), , drop = FALSE]
        b <- qr.coef(qr = qr(segx), y = segy)
        res <- segy - segx %*% b
        umat <- reshapex(res, nrow(res)/n, n)
        vvar <- crossprod(x = umat, y = NULL) / (j - i + 1)
        vstar <- vvar + 1
        itr <- 1L
        while ((maxc(abs(vec(t(vvar - vstar)))) > 1E-6) && (itr < max_iter)) {
            vstar <- vvar
            ibigv <- eye(nrow(segy)/n) %x% qr.solve(vstar)
            b <- qr.solve(t(segx) %*% ibigv %*% segx, t(segx) %*% ibigv %*% segy)
            res <- segy - segx %*% b
            umat <- reshapex(res, nrow(res)/n, n)
            vvar <- crossprod(x = umat, y = NULL) / (j - i + 1)
            itr <- itr + 1L
            if (itr == max_iter) {
                warning("FGLS failed to converge\n", call. = FALSE)
                break
            }
        }
        
        bhat[k, ] <- t(b)
        vv[((k - 1) * n + 1):(k * n), ] <- vvar
        k <- k + 1L
    }

    return (list(bhat = bhat, vv = vv))

}

# Estimate the model allowing for cross regime (linear) restrictions
restim <- function(maty, matz, bigt, n, m, br, R, brbeta, brv, max_iter = 1000) {

# Inputs:
#   maty = dependent variable
#   matz = independent variable(s)
#   n = number of equations
#   m = number of breaks
#   bigt = sample size
#   br = break dates
#   R = the restriction, taking the form beta = R %*% teta, where teta is the vector of basic 
#       parameters of the restricted model.
#   brv:
#       brv == 1 if the coefficients are allowed to change
#       brv == 0 otherwise
#   brbeta:
#       brbeta == 1 if the coefficients are allowed to change
#       brbeta == 0 otherwise
#
# Output (list object):
#   nbeta = the estimate of coefficients imposing restrictions
#   nvv = the estimate of covariances imposing restrictions
#
# Note:
#   the model includes partial structural change and block-partial structural change models

seg <- rbind(0, br, (nrow(maty) / n))

# Now estimate an unrestricted model using estim()

tmp <- estim(maty, matz, n, m, br)
beta <- tmp$bhat
vv <- tmp$vv

# Changes in both coefficients and covariances
if ((brv == 1) && (brbeta == 1)) {
    pmatz <- pzbar(matz, m, br, bigt) # %*% R
    pmatz <- pmatz %*% R
    k <- 1L
    ibigv <- eye(bigt * n)
    while (k <= (m + 1)) {
        i <- seg[k] + 1
        j <- seg[(k + 1)]
        ibigv[((i - 1)*n+1):(j * n), ((i - 1)*n+1):(j * n)] <- eye(j - i + 1) %x% qr.solve(vv[((k - 1)*n+1):(k * n), ])
        k <- k + 1L
    }

    b <- qr.solve(t(pmatz) %*% ibigv %*% pmatz, t(pmatz) %*% ibigv %*% maty)
    bstar <- b + 10
    itr <- 1L
    while ((maxc(abs(bstar - b)) > 1E-6) && (itr < max_iter)) {
        bstar <- b
        k <- 1L
        # Update the VCV matrix
        while (k <= (m + 1)) {
            i <- seg[k] + 1
            j <- seg[k + 1]
            tmpy <- maty[((i - 1)*n+1):(j * n), , drop = FALSE]
            tmpx <- pmatz[((i - 1)*n+1):(j * n), , drop = FALSE]
            res <- tmpy - tmpx %*% bstar
            umat <- reshapex(res, nrow(res)/n, n)
            vvar <- crossprod(x = umat, y = NULL) / (j - i + 1)
            vv[((k - 1)*n+1):(k * n), ] <- vvar
            k <- k + 1L
        }
        
        k <- 1L
        # Create the block diagonal matrix of the covariances
        while (k <= (m + 1)) {
            i <- seg[k] + 1
            j <- seg[(k + 1)]
            ibigv[((i - 1)*n+1):(j * n), ((i - 1)*n+1):(j * n)] <- eye(j - i + 1) %x% qr.solve(vv[((k - 1)*n+1):(k * n), ])
            k <- k + 1L
        }
        
        b <- qr.solve(t(pmatz) %*% ibigv %*% pmatz, t(pmatz) %*% ibigv %*% maty)
        itr <- itr + 1L
        if (itr == max_iter) {
            warning("Iteration has reached the upper bound.\n", call. = FALSE)
            break
        }
    }

    nbeta <- R %*% b
    nvv <- vv
}

# Changes in covariances only
# iterative method is used for the optimisation, with the unrestricted estimate as the prior input.
# The procedure is iterated until convergence.
if ((brv == 1) && (brbeta != 1)) {
    ibigv <- eye(bigt * n)
    k <- 1L
    while (k <= (m + 1)) {
        i <- seg[k] + 1
        j <- seg[(k + 1)]
        ibigv[((i - 1)*n+1):(j * n), ((i - 1)*n+1):(j * n)] <- eye(j - i + 1) %x% qr.solve(vv[((k - 1)*n+1):(k * n), ])
        k <- k + 1L
    }

    b <- qr.solve(t(matz) %*% ibigv %*% matz, t(matz) %*% ibigv %*% maty)
    bstar <- b + 10
    itr <- 1L
    while ((maxc(abs(bstar - b)) > 1E-6) && (itr < max_iter)) {
        bstar <- b
        k <- 1L
        # Update the VCV matrix
        while (k <= (m + 1)) {
            i <- seg[k] + 1
            j <- seg[k + 1]
            tmpy <- maty[((i - 1)*n+1):(j * n), , drop = FALSE]
            tmpx <- matz[((i - 1)*n+1):(j * n), , drop = FALSE]
            res <- tmpy - tmpx %*% bstar
            umat <- reshapex(res, nrow(res)/n, n)
            vvar <- crossprod(x = umat, y = NULL) / (j - i + 1)
            vv[((k - 1)*n+1):(k * n), ] <- vvar
            k <- k + 1L
        }
        
        k <- 1L
        # Create the block diagonal matrix of the covariances
        while (k <= (m + 1)) {
            i <- seg[k] + 1
            j <- seg[(k + 1)]
            ibigv[((i - 1)*n+1):(j * n), ((i - 1)*n+1):(j * n)] <- eye(j - i + 1) %x% qr.solve(vv[((k - 1)*n+1):(k * n), ])
            k <- k + 1L
        }
        
        b <- qr.solve(t(matz) %*% ibigv %*% matz, t(matz) %*% ibigv %*% maty)
        itr <- itr + 1L
        if (itr == max_iter) {
            warning("Iteration has reached the upper bound.\n", call. = FALSE)
            break
        }
    }
        
    nbeta <- R %*% b
    nvv <- vv
    k <- 2L
    while (k <= (m + 1)) {
        nbeta <- rbind(nbeta, b)
        k <- k + 1L
    }
}
    
# Changes in coefficients only
if ((brv == 1) && (brbeta != 1)) {
    k <- 1L
    vvar <- 0
    while (k <= (m + 1)) {
        i <- seg[k] + 1
        j <- seg[k + 1]
        tmpy <- maty[((i - 1)*n+1):(j * n), , drop = FALSE]
        tmpx <- matz[((i - 1)*n+1):(j * n), , drop = FALSE]
        res <- tmpy - tmpx %*% t(beta[k, , drop = FALSE])
        umat <- reshapex(res, nrow(res)/n, n)
        vvar <- vvar + crossprod(x = umat, y = NULL)
        k <- k + 1L
    }
    
    vvar <- vvar / bigt
    ibigv <- eye(bigt) %x% qr.solve(vvar)
    pmatz <- pzbar(matz, m, br, bigt)
    b <- qr.solve(t(pmatz) %*% ibigv %*% pmatz, t(pmatz) %*% ibigv %*% maty)
    # Now iterate until convergence is reached
    bstar <- b + 10
    itr <- 1L
    while ((maxc(abs(bstar - b)) > 1E-6) && (itr < max_iter)) {
        bstar <- b
        k <- 1L
        res <- maty - pmatz %*% bstar
        umat <- reshapex(res, nrow(res)/n, n)
        vvar <- crossprod(x = umat, y = NULL) / bigt
        ibigv <- eye(bigt) %x% qr.solve(vvar)
        b <- qr.solve(t(pmatz) %*% ibigv %*% pmatz, t(pmatz) %*% ibigv %*% maty)
        itr <- itr + 1
        if (itr == max_iter) {
            warning("Iteration has reached the upper bound.\n", call. = FALSE)
            break
        }
    }
    
    nbeta <- R %*% b
    nvv <- vvar
    k <- 2L
    while (k <= (m + 1)) {
        nvv <- rbind(nvv, vvar)
        k <- k + 1L
    }
}

# return estimated coefficients and covariances
return (list(nbeta = nbeta, nvv = nvv))

}

# Construct the diagonal partition of matz with m breaks at dates bb
pzbar <- function(matz, m, bb, bigt) {

    # Get some useful values
    nt <- nrow(matz)
    q1 <- ncol(matz)
    n <- nt / bigt
    
    if (m == 0) {
        stop("m == 0, no break is allowed.\n")
    } else {
        matzb <- matrix(data = 0, nrow = nt, ncol = (m + 1) * q1)
        matzb[1:(bb[1] * n), 1:q1] <- matz[1:(bb[1] * n), ]
        i <- 2L
        while (i <= m) {
            matzb[(bb[i - 1]*n+1):(bb[i] * n), ((i - 1)*q1+1):(i * q1)] <- matz[(bb[i - 1]*n+1):(bb[i] * n), ]
            i <- i + 1L
        }
        
        matzb[(bb[m]*n+1):nt, (m*q1+1):((m + 1) * q1)] <- matz[(bb[m]*n+1):nt, ]
    }
    return (matzb)
}    

# Estimate robust standard errors
correct <- function(res, prewhit) {
    
    # Get some useful values
    d <- ncol(res)
    nt <- nrow(res)
    bhat <- matrix(data = 0.0, nrow = d, ncol = 1)
    bmat <- matrix(data = 0.0, nrow = d, ncol = d)
    vstar <- matrix(data = 0.0, nrow = (nt - 1), ncol = d)
    vmat <- res
    
    # Apply pre-whitening?
    if (prewhit) {
        for (i in 1L:d) {
            bhat <- qr.coef(qr = qr(vmat[1:(nt - 1), , drop = FALSE]), y = vmat[2:nt, i, drop = FALSE])
            bmat[, i] <- bhat
            vstar[, i] <- vmat[2:nt, i, drop = FALSE] - vmat[1:(nt - 1), , drop = FALSE] %*% bhat
        }
    
        # Call the kernel on the residuals
        jh <- jhatpr(vstar)
        
        # Re-colour
        I_bhat_i <- qr.solve(eye(d) - bmat)
        hac <- I_bhat_i %*% jh %*% t(I_bhat_i)
    } else {
        hac <- jhatpr(vmat)
    }
    return (hac)
}

# Compute long-run VCV matrix of vmat
jhatpr <- function(vmat) {

    # Get a few useful values
    nt <- nrow(vmat)
    d <- ncol(vmat)
    jhat <- matrix(data = 0.0, nrow = d, ncol = d)
    
    # Call the automatic bandwidth selector
    st <- bandw(vmat)
    
    # HAC correction: Lag 0
    jhat <- crossprod(x = vmat, y = NULL) # vmat'vmat
    
    # HAC correction: Lag 1,...,T-1
    # Forward summation...
    for (j in 1L:(nt - 1)) { 
        jhat <- jhat + kern(j/st) * crossprod(x = vmat[(j + 1):nt, ], y = vmat[1:(nt - j), ])
    }
    # ...and Backward summation.
    for (j in 1L:(nt - 1)) {
        jhat <- jhat + kern(j/st) * crossprod(x = vmat[1:(nt - j), ], y = vmat[(j + 1):nt, ])
    }
    
    # Do small sample correction as specified by Andrews (1991)
    jhat <- jhat / (nt - d)
    return (jhat)
}

# Compute Andrews (1991) automatic bandwidth using and AR(1) estimated via OLS
bandw <- function(vhat) {

    # Get some useful values
    nt <- nrow(vhat)
    d <- ncol(vhat)
    a2n <- 0.0
    a2d <- 0.0
    
    for (j in 1L:d) {
        v <- vhat[, j, drop = FALSE]
        vt <- v[2:nt]
        vt_1 <- v[1:(nt - 1)]
        rho_j <- sum(vt_1 * vt) / sum(vt_1 * vt_1); # NB: We use ||x'y|| / ||x'x|| for speed
        et <- vt - (rho_j * vt_1);
        s2 <- sum(et * et) / (nt - 1);
        a2n <- a2n + ((4 * rho_j * rho_j * s2 * s2) / (1 - rho_j)^8)
        a2d <- a2d + ((s2 * s2) / (1 - rho_j)^4)
    }
    
    a2 <-a2n / a2d
    st <- 1.3221 * ((a2 * nt)^0.2)
    return (st)
}

# Compute Quadratic Spectral Kernel
kern <- function(x) {
    del <- (6.0 * pi * x) / 5.0
    ker <- 3.0 * (sin(del) / del - cos(del)) / (del * del)
    return (ker)
}

make_br_vec <- function(x)
{
    return (matrix(data = x, nrow = length(x)))
}

# Function to print out confidence sets information (length and included observation indices)
print_csm_info <- function(csm, level) {
    
    m <- ncol(csm)
    bigt <- nrow(csm)    
    
    cat("\n-------------------------------------------------------------------------\n")
    cat(sprintf("Confidence Set Level = %d%%\n\n", level))
    for(i in 1L:m) {
        csm_idx <- which(!is.na(csm[, i]), arr.ind = TRUE) 
        cat(sprintf("Break %d confidence set's length is %d observations and includes:\n", 
            i, length(csm_idx)))
        print.default(csm_idx)
        cat('\n')
    }

}

# Print method for class "invLR"
print.invLR <- function(x, digits = max(3L, getOption("digits") - 3L), print_all = TRUE, ...) {
	cat("\nCall: ", paste(deparse(x$call), sep = "\n", collapse = "\n"), "\n", sep = "")
    
    cat("\n\tInverted LR Confidence Sets (Eo and Morley 2015):\n")
    
    cat("\nConfidence Sets:\n")
    print_csm_info(x$csm99, level = 99)
    
    print_csm_info(x$csm95, level = 95)
    
    print_csm_info(x$csm90, level = 90)
    
    if (print_all) {
        cat("\nCritical Values (kappa):\n")
        print.default(x$kmat)
        
        cat("\nRestricted Beta:\n")
        print.default(x$rbetast)
        
        cat("\nRestricted Variance:\n")
        print.default(x$rvvst)
    }
    cat('\n')
}

# Plot method for class "invLR"
plot.invLR <- function(x, col_y = "black", lwd_y = 1, lwd_cs = 3, lwd_br = 2, col_af = 0.1) {
    
    # Set up plotting device
    op <- par(mar = c(3, 3, 2, 1) + 0.1, oma = c(0, 0, 3, 0), bty = 'o', tcl = -0.5, 
        lty = 1, lwd = 1, xaxs = 'r', yaxs = 'r', mgp = c(3, 1, 0), 
        cex.main = 1.2, font.main = 2, cex.lab = 1.1, font.lab = 1, cex.axis = 1.1, font.axis = 1)
    
    on.exit(par(op))
    
    m <- ncol(x$csm99)
    y_names <- colnames(x$y)
    if(is.null(y_names)) {
        y_names <- paste0('Y', 1:ncol(x$y))
    }
    delta <- 1.5

    layout(matrix(1L:(3L * ncol(x$y)), nrow = 3L, ncol = ncol(x$y), byrow = FALSE))
    
    for (i in 1L:ncol(x$y)) {
        
        # Gets upper and lower bounds
        ub <- max(x$y[, i]) * delta
        lb <- min(x$y[, i]) * delta
    
        # 99% Confidence set
        plot(x$y[, i], type = 'n', col = NA, lty = 1, lwd = 1, 
            main = sprintf("99%% Level - %s", y_names[i]), ylab = "", xlab = "")
        
        # Confidence set markers and estimated break point
        for (j in 1L:m) {
            lines(ifelse(x$csm99[, j] == 1, ub, x$csm99[, j]), type = 'h', 
                col = adjustcolor(j+1, alpha.f = col_af), 
                lwd = lwd_cs, lend = "square", ylim = c(lb, ub))
            lines(ifelse(x$csm99[, j] == 1, lb, x$csm99[, j]), type = 'h',
                col = adjustcolor(j+1, alpha.f = col_af), 
                lwd = lwd_cs, lend = "square", ylim = c(lb, ub))
            abline(v = x$br[j], col = j+1, lty = 1, lwd = lwd_br)
        }
        
        abline(h = 0, col = "black", lty = 1, lwd = 1)
        lines(x$y[, i], type = 'l', col = col_y, lty = 1, lwd = lwd_y)
        box(which = "plot", col = "black", lty = 1, lwd = 1)
        
        # 95% Confidence set
        plot(x$y[, i], type = 'n', col = NA, lty = 1, lwd = 1, 
            main = sprintf("95%% Level - %s", y_names[i]), ylab = "", xlab = "")
        
        # Confidence set markers and estimated break point
        for (j in 1L:m) {
            lines(ifelse(x$csm95[, j] == 1, ub, x$csm95[, j]), type = 'h',
                col = adjustcolor(j+1, alpha.f = col_af),
                lwd = lwd_cs, lend = "square", ylim = c(lb, ub))
            lines(ifelse(x$csm95[, j] == 1, lb, x$csm95[, j]), type = 'h', 
                col = adjustcolor(j+1, alpha.f = col_af),
                lwd = lwd_cs, lend = "square", ylim = c(lb, ub))
            abline(v = x$br[j], col = j+1, lty = 1, lwd = lwd_br)
        }
        
        abline(h = 0, col = "black", lty = 1, lwd = 1)        
        lines(x$y[, i], type = 'l', col = col_y, lty = 1, lwd = lwd_y)
        box(which = "plot", col = "black", lty = 1, lwd = 1)
        
        # 90% Confidence set
        plot(x$y[, i], type = 'n', col = NA, lty = 1, lwd = 1, 
            main = sprintf("90%% Level - %s", y_names[i]), ylab = "", xlab = "")
        
        # Confidence set markers and estimated break point
        for (j in 1L:m) {
            lines(ifelse(x$csm90[, j] == 1, ub, x$csm90[, j]), type = 'h', 
                col = adjustcolor(j+1, alpha.f = col_af),
                lwd = lwd_cs, lend = "square", ylim = c(lb, ub))
            lines(ifelse(x$csm90[, j] == 1, lb, x$csm90[, j]), type = 'h',
                col = adjustcolor(j+1, alpha.f = col_af),
                lwd = lwd_cs, lend = "square", ylim = c(lb, ub))
            abline(v = x$br[j], col = j+1, lty = 1, lwd = lwd_br)
        }
        
        abline(h = 0, col = "black", lty = 1, lwd = 1)        
        lines(x$y[, i], type = 'l', col = col_y, lty = 1, lwd = lwd_y)
        box(which = "plot", col = "black", lty = 1, lwd = 1)
        
    }
    
    # Main plot title
    mtext("Inverted LR Confidence Sets", side = 3, outer = TRUE, cex = 1.2, font = 2)
}

# EOF
